/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  3dnow.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/02/27 10:16:45 $
 * $Log: 3dnow.c,v $
 * Revision 1.3.2.1  2002/02/27 10:16:45  jmaebe
 * * prefixes are counted in ORIGINAL_INSTRUCTION -> pass 3 instead of 4
 *   to no_mem for reg-reg 3dnow instructions
 *
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"


extern unsigned char byte1, byte2, byte3, byte4, prefix_c, immsize_override;


unsigned _3dnow(unsigned dummy1, unsigned dummy2, unsigned dummy3){
  unsigned char mod2=byte2>>6;

  /* 
     Hier komen we terecht voor alle 3dnow-instructies, uitgezonderd
     FEMMS en PREFETCH (zijn nooit geheugenoperaties). Deze
     laatste worden direct door no_mem() afgehandeld. 
  */


  // In feite moeten we 2 0F-prefixen skippen en hebben we dus nog 1 skip() nodig.
  // Echter: de opcode staat niet vooraan!!! Daarom laten we de 2e OF-prefix (als een
  // dummy opcode) staan  om check_mod() te kunnen gebruiken.


  if (mod2==3) // Fast exit...

    return no_mem(3 /*instr_len*/, 0 /*dummy*/, 0 /*dummy*/);   // reg,reg

  else
    
    // 3DNow kent enkel 64bit leesoperaties => mem_len=8, type=0

    // Merk op dat een 3DNow nooit een immediate heeft! Toch zetten we imm_len=1 omdat dit
    // de enige manier is om de extra 0F-prefix te laten meetellen!

    return check_mod(1 /*imm_len*/, 8 /*mem_len*/, LOAD /*type*/);



}
