//#define DEBUG

/*
pthread_rwlock !!!
*/


/*
POSIX 1003.1b semaphores
 */


#include <pthread.h>
#include <semaphore.h>
#include <errno.h>
#include "jiti-dr.h"

extern segment_info _MR_segment_now[MAXTHREADS];

vct * get_Object_VCP(void * address);

// Debugprocedure
#define PRINTVC(VCP)  _MR_print_vc(VCP)
#define PRINTVCT      PRINTVC(&(_MR_segment_now[THREADID].vc))
#define DPRINTVCT     DBG(PRINTVCT)

///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///
   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///   
///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///

#define AANTAL_INSTR_FUNC    10

#define PTHREAD_EXIT          0
#define PTHREAD_JOIN          1
#define PTHREAD_MUTEX_LOCK    2
#define PTHREAD_MUTEX_UNLOCK  3
#define PTHREAD_MUTEX_TRYLOCK 4

#define SEM_WAIT              5
#define SEM_POST              6
#define SEM_TRYWAIT           7

#define COND_WAIT             8
#define COND_TIMEDWAIT        9

int dr_pthread_join(pthread_t th, void **thread_return);
void dr_pthread_exit(void *retval);

int dr_pthread_mutex_lock(pthread_mutex_t *mutex);
int dr_pthread_mutex_unlock(pthread_mutex_t *mutex);
int dr_pthread_mutex_trylock(pthread_mutex_t *mutex);

int dr_sem_wait(sem_t * sem);
int dr_sem_trywait(sem_t * sem);
int dr_sem_post(sem_t * sem);

int dr_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
int dr_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *);


///
/// RELOC INFO
///

struct _reloc_info_ {

  char function[50];
  unsigned real_address;
  unsigned instr_address;

} reloc_info[AANTAL_INSTR_FUNC] = {

  {"pthread_exit",            0, (unsigned)dr_pthread_exit},           // 0
  {"pthread_join",            0, (unsigned)dr_pthread_join},           // 1
  

  {"pthread_mutex_lock",      0, (unsigned)dr_pthread_mutex_lock},     // 2
  {"pthread_mutex_unlock",    0, (unsigned)dr_pthread_mutex_unlock},   // 3
  {"pthread_mutex_trylock",   0, (unsigned)dr_pthread_mutex_trylock},  // 4

  {"sem_wait",                0, (unsigned)dr_sem_wait},               // 5
  {"sem_post",                0, (unsigned)dr_sem_post},               // 6
  {"sem_trywait",             0, (unsigned)dr_sem_trywait},            // 7

  {"pthread_cond_wait",       0, (unsigned)dr_pthread_cond_wait},      // 8
  {"pthread_cond_timedwait",  0, (unsigned)dr_pthread_cond_timedwait}, // 9

};


///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///
   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///   
///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///


/////////////////////////////////////////////////////////////////
// done /////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////

done(){
  /* 't is gedaan, sluit laatste segment v/ draad 1 af */
  _MR_segment_switch(0, 0);

  // indien we hier geraken is het gedaan en waren er geen dataraces
  printf("No data race found.\n");

}

////////////////////////////////////////////////////////////////
// mem /////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

mem(unsigned ip, unsigned type, unsigned len, unsigned address){  
  
  // We behandelen een MODIFY als een STORE, dit is OK voor DR-detectie.
  if (type==MODIFY) type=STORE;

  _MR_add_to_bitmap(address, len, THREADID, type);
  
}

////////////////////////////////////////////////////////////////
// CREATE //////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

struct thread_info_t {
  void * (*start_routine)(void *);
  void * arg;
  vct vc_father;
};

unsigned dr_real_pthread_create;

void dr_pthread_create_helper(struct thread_info_t * thread_info){
  
  // start first segment using VC of father
  _MR_segment_switch(&(thread_info->vc_father), 0);

  DBG(printf("THIS IS THREAD %d/%d, my VC is ", THREADID, pthread_self())); DBG(PRINTVCT) ;

  // start execution
  (*thread_info->start_routine)(thread_info->arg);

  // end of thread
  free(thread_info);
  
}

///////////////////////////////////////////////////////////////////

int dr_pthread_create(pthread_t * thread, const pthread_attr_t * attr, 
		      void * (*start_routine)(void *), void * arg){
  int status;
  struct thread_info_t * thread_info;
  
  // switch to next segment
  _MR_segment_switch(0, 0);

  DBG(printf("T%d: pthread_create\n", THREADID));

  DBG(printf("THIS IS THE FATHER  %d/%d, my VC is ", THREADID, pthread_self())); DBG(PRINTVCT) ;

  thread_info=(struct thread_info_t *)malloc(sizeof(struct thread_info_t));

  // LET OP !!!!  +delta  !!!!
  thread_info->start_routine = start_routine+delta;
  thread_info->arg = arg;
  memcpy(thread_info->vc_father, _MR_segment_now[THREADID].vc, sizeof(vct));

  // LET OP  !!!!  -delta  !!!!   
  status=(*(int(*)(pthread_t*, const pthread_attr_t*, void*(*)(void*), void*))dr_real_pthread_create)
    (thread, attr, (void*(*)(void*))dr_pthread_create_helper-delta,thread_info);

  return status;
}

/////////////////////////////////////////////////////////////////
// JOIN /////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////

int dr_pthread_join(pthread_t th, void **thread_return){
  int status;

  DBG(printf("T%d: +pthread_join with ", THREADID, TID(th))); DBG(PRINTVCT);

  // join
  status=(*(int(*)(pthread_t, void**))reloc_info[PTHREAD_JOIN].real_address)(th, thread_return);

  // start nieuw segment
  _MR_segment_switch(&(_MR_segment_now[TID(th)].vc), 0);

  DBG(printf("T%d: -pthread_join ", THREADID)); DBG(PRINTVCT);

  return status;
}

/////////////////////////////////////////////////////////////////
// EXIT /////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////

void dr_pthread_exit(void *retval){

  DBG(printf("T%d: +pthread_exit ", THREADID)); DBG(PRINTVCT);

  // beeindig laatste segment
  _MR_segment_switch(0, 0);

  DBG(printf("T%d: -pthread_exit ", THREADID)); DBG(PRINTVCT);

  // beeindig draad
  (*(void(*)(void*))reloc_info[PTHREAD_EXIT].real_address)(retval);

  // not reached
}


//////////////////////////////////////////////////////////////////
// MUTEX_LOCK ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_pthread_mutex_lock(pthread_mutex_t *mutex){
  int status;
  
  DBG(printf("T%d: +pthread_mutex_lock(%x)\n", THREADID, mutex));

  status = (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_LOCK].real_address)(mutex);

  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);  

  DBG(printf("T%d: -pthread_mutex_lock(%x)\n", THREADID, mutex));

  return status;
}

//////////////////////////////////////////////////////////////////
// MUTEX_TRYLOCK ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_pthread_mutex_trylock(pthread_mutex_t *mutex){
  int status = (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_TRYLOCK].real_address)(mutex);

  DBG(printf("T%d: pthread_mutex_trylock(%x)\n", THREADID, mutex));

  // did we get the lock?
  if (status==EBUSY) return EBUSY;  // NO

  // we have the lock, proceed as pthread_mutex_lock()
  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);
  
  return 0;

}

//////////////////////////////////////////////////////////////////
// MUTEX_UNLOCK //////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_pthread_mutex_unlock(pthread_mutex_t *mutex){

  DBG(printf("T%d: +pthread_mutex_unlock(%x)\n", THREADID, mutex));

  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);
  
  DBG(printf("T%d: -pthread_mutex_unlock(%x)\n", THREADID, mutex));

  return (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_UNLOCK].real_address)(mutex);

}

//////////////////////////////////////////////////////////////////
// SEM_WAIT //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_sem_wait(sem_t * sem){
  int status = (*(int(*)(sem_t*))reloc_info[SEM_WAIT].real_address)(sem);

  DBG(printf("T%d: sem_wait\n", THREADID));

  _MR_segment_switch(get_Object_VCP(sem), UPDATE_VC_O | LOCK_VC_O);
  
  return status;
}

//////////////////////////////////////////////////////////////////
// SEM_POST //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_sem_post(sem_t * sem){

  DBG(printf("T%d: sem_post\n", THREADID));

  _MR_segment_switch(get_Object_VCP(sem), UPDATE_VC_O | LOCK_VC_O);
  
  return (*(int(*)(sem_t*))reloc_info[SEM_POST].real_address)(sem);
}

//////////////////////////////////////////////////////////////////
// SEM_TRYWAIT ///////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_sem_trywait(sem_t * sem){
  int status = (*(int(*)(sem_t*))reloc_info[SEM_TRYWAIT].real_address)(sem);

  DBG(printf("T%d: sem_trywait\n", THREADID));

  // Hebben we de semafoor?
  if (status==EAGAIN) return EAGAIN;   // NEE

  _MR_segment_switch(get_Object_VCP(sem), UPDATE_VC_O | LOCK_VC_O );

  return status;
}

//////////////////////////////////////////////////////////////////
// COND_WAIT /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex){
  int status;

  DBG(printf("T%d: +pthread_cond_wait(%x, %x)\n", THREADID, cond, mutex));

  // switch voor interne mutex_unlock. Merk op dat we de mutex nog hebben.
  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);

  status = (*(int(*)(pthread_cond_t*, pthread_mutex_t*))reloc_info[COND_WAIT].real_address)(cond, mutex);

  // switch voor interne mutex_lock. Merk op dat we de mutex reeds terug hebben.
  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);

  DBG(printf("T%d: -pthread_cond_wait(%x, %x)\n", THREADID, cond, mutex));

  return status;
}

//////////////////////////////////////////////////////////////////
// COND_TIMEDWAIT ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int dr_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec * abstime){
  int status;

  DBG(printf("T%d: pthread_cond_timedwait\n", THREADID));

  // switch voor interne mutex_unlock. Merk op dat we de mutex nog hebben.
  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);

  status = (*(int(*)(pthread_cond_t*, pthread_mutex_t*, const struct timespec*))
	    reloc_info[COND_TIMEDWAIT].real_address)(cond, mutex, abstime);

  // switch voor interne mutex_lock. Merk op dat we de mutex reeds terug hebben, ook al was er een timeout!
  _MR_segment_switch(get_Object_VCP(mutex), UPDATE_VC_O);

  return status;
}

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

backend_init(){
  int i=0;

  fprintf(jiti_tty, "** Data race detection.\n");

  if (dr_real_pthread_create = jiti_callback("pthread_create", (unsigned)dr_pthread_create)) {
    
    // PARALLEL PROGRAM
    fprintf(jiti_tty, "** Parallel program detected. OK.\n");

    // onderschep alle synchronisatiefuncties
    while (i<AANTAL_INSTR_FUNC){
      reloc_info[i].real_address = jiti_callback(reloc_info[i].function, reloc_info[i].instr_address);
      i++;
    }

    // onderschep het beeindigen v/h programma
    jiti_callback("#STOP#", (unsigned)done);

    // onderschep alle geheugenoperaties
    jiti_callback("#DATA#",  (unsigned)mem);
    
    /* start het eerste segment v/d eerste draad */
    _MR_segment_switch(0, 0);
 
  } else {
    // SEQUENTIAL PROGRAM
    fprintf(jiti_tty, "** This is a sequential program! Bailing out...\n");
  }

}

//////////////////////////////////////////////////////////////////
//INTERNE functies ///

int _MR_INTERN_mutex_lock(pthread_mutex_t * mutex){
  return reloc_info[PTHREAD_MUTEX_LOCK].real_address ?
    (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_LOCK].real_address)(mutex) :
    pthread_mutex_lock(mutex) ;
}

int _MR_INTERN_mutex_unlock(pthread_mutex_t * mutex){
  return reloc_info[PTHREAD_MUTEX_UNLOCK].real_address ? 
    (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_UNLOCK].real_address)(mutex) :
    pthread_mutex_unlock(mutex) ;
}

//////////////////////////////////////////////////////////////////

INIT(backend_init);

///////////////////////

#define T_SIZE 100

struct _O_2_VCP_ {
  unsigned address;
  unsigned vcp;
} O_2_VCP[T_SIZE];


vct * get_Object_VCP(void * address){
  int i=0, empty_spot=-1;


  while (i<T_SIZE){

    // found address
    if (O_2_VCP[i].address==(unsigned)address)  return (vct*)O_2_VCP[i].vcp;

    // found empty spot
    if ((O_2_VCP[i].address==0) && (empty_spot==-1)) empty_spot=i;
    
    i++;
  }

  // not found
  //  case 1: table full
  if (empty_spot==T_SIZE) panic("O_2_VCP is full!\n");
  
  //  case 2: empty spot at i
  O_2_VCP[empty_spot].address = (unsigned)address;
  O_2_VCP[empty_spot].vcp     = calloc(1, sizeof(vct));

  return (vct*)O_2_VCP[empty_spot].vcp;
}
