//#define DEBUG

#include "jiti-recplay.h"

LCt * recplay_get_Object_LCP(void * address);

////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
                                                                                     ///
#ifndef RECORD                                                                       ///
#ifndef REPLAY                                                                       ///
#error You should either define RECORD or REPLAY!                                    ///
#endif                                                                               ///
#endif                                                                               ///
                                                                                     ///
#ifdef RECORD                                                                        ///
#define ifREC(X) X                                                                   ///
#else                                                                                ///
#define ifREC(X)                                                                     ///
#endif                                                                               ///
                                                                                     ///
#ifdef REPLAY                                                                        ///
#define ifREP(X) X                                                                   ///
#else                                                                                ///
#define ifREP(X)                                                                     ///
#endif                                                                               ///
                                                                                     ///
#define ifREC_UPDATE_LC(address) ifREC(_MR_calc_lt(recplay_get_Object_LCP(address))) ///
#define ifREP_WAIT()             ifREP(_MR_replay_try())                             ///
#define ifREP_NEXT()             ifREP(_MR_replay_next())                            ///
                                                                                     ///
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////


/*
pthread_rwlock !!!
*/


/*
POSIX 1003.1b semaphores
 */



// Debugprocedure
#define PRINTLC(LC)   printf("lc=%d\n", LC)
#define PRINTLCT      PRINTLC(_MR_thr_LC[THREADID])
#define DPRINTLCT     DBG(PRINTLCT)

///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///
   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///   
///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///

#define AANTAL_INSTR_FUNC    10

#define PTHREAD_EXIT          0
#define PTHREAD_JOIN          1
#define PTHREAD_MUTEX_LOCK    2
#define PTHREAD_MUTEX_UNLOCK  3
#define PTHREAD_MUTEX_TRYLOCK 4

#define SEM_WAIT              5
#define SEM_POST              6
#define SEM_TRYWAIT           7

#define COND_WAIT             8
#define COND_TIMEDWAIT        9

int  recplay_pthread_join(pthread_t th, void **thread_return);
void recplay_pthread_exit(void *retval);

int recplay_pthread_mutex_lock(pthread_mutex_t *mutex);
int recplay_pthread_mutex_unlock(pthread_mutex_t *mutex);
int recplay_pthread_mutex_trylock(pthread_mutex_t *mutex);

int recplay_sem_wait(sem_t * sem);
int recplay_sem_trywait(sem_t * sem);
int recplay_sem_post(sem_t * sem);

int recplay_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex);
int recplay_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, const struct timespec *);


///
/// RELOC INFO
///

struct _reloc_info_ {

  char function[50];
  unsigned real_address;
  unsigned instr_address;

} reloc_info[AANTAL_INSTR_FUNC] = {

  {"pthread_exit",            0, (unsigned)recplay_pthread_exit},           // 0
  {"pthread_join",            0, (unsigned)recplay_pthread_join},           // 1
  

  {"pthread_mutex_lock",      0, (unsigned)recplay_pthread_mutex_lock},     // 2
  {"pthread_mutex_unlock",    0, (unsigned)recplay_pthread_mutex_unlock},   // 3
  {"pthread_mutex_trylock",   0, (unsigned)recplay_pthread_mutex_trylock},  // 4

  {"sem_wait",                0, (unsigned)recplay_sem_wait},               // 5
  {"sem_post",                0, (unsigned)recplay_sem_post},               // 6
  {"sem_trywait",             0, (unsigned)recplay_sem_trywait},            // 7

  {"pthread_cond_wait",       0, (unsigned)recplay_pthread_cond_wait},      // 8
  {"pthread_cond_timedwait",  0, (unsigned)recplay_pthread_cond_timedwait}, // 9

};


///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///
   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///   
///   ///   ///   ///   ///   ///   ///   ///   ///   ///   ///


/////////////////////////////////////////////////////////////////
// done /////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////

done(){
  fprintf(jiti_tty, "===== Recplay: done\n");

  ifREC(_MR_stop_record());
  ifREP(_MR_stop_replay());
    
}

////////////////////////////////////////////////////////////////
// CREATE //////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

struct thread_info_t {
  void * (*start_routine)(void *);
  void * arg;
  LCt start_lc;
};

unsigned recplay_real_pthread_create;

void recplay_pthread_create_helper(struct thread_info_t * thread_info){

  // startwaarde voor LC
  ifREC( _MR_thr_first_LC[ THREADID ] = _MR_thr_LC[ THREADID ] = thread_info->start_lc );

  DBG(printf("THIS IS THREAD %d/%d, my LC is ", THREADID, pthread_self())); DBG(PRINTLCT);

  // start execution
  (*thread_info->start_routine)(thread_info->arg);

  // end of thread
  free(thread_info);

}

///////////////////////////////////////////////////////////////////

int recplay_pthread_create(pthread_t * thread, const pthread_attr_t * attr, 
		      void * (*start_routine)(void *), void * arg){
  int status;
  struct thread_info_t * thread_info;
  
  DBG(printf("T%d: +pthread_create\n", THREADID));

  thread_info=(struct thread_info_t *)malloc(sizeof(struct thread_info_t));

  // advance our LC
  ifREC_UPDATE_LC(NULL);

  DBG(printf("THIS IS THE FATHER  %d/%d, my LC is ", THREADID, pthread_self())); DBG(PRINTLCT) ;

  ifREP_WAIT();

  // LET OP !!!!  +delta  !!!!
  thread_info->start_routine  = start_routine+delta;
  thread_info->arg            = arg;
  ifREC(thread_info->start_lc = _MR_thr_LC[ THREADID ]);

  // LET OP  !!!!  -delta  !!!!   
  status=(*(int(*)(pthread_t*, const pthread_attr_t*, void*(*)(void*), void*))recplay_real_pthread_create)
    (thread, attr, (void*(*)(void*))recplay_pthread_create_helper-delta,thread_info);
  
  ifREP_NEXT();

  DBG(printf("T%d: -pthread_create\n", THREADID));

  return status;
}

/////////////////////////////////////////////////////////////////
// JOIN /////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////

int recplay_pthread_join(pthread_t th, void **thread_return){
  int status;
  
  DBG(printf("T%d: +pthread_join with %d ", THREADID, TID(th))); DBG(PRINTLCT);
  
  ifREP_WAIT();
  
  // join
  status=(*(int(*)(pthread_t, void**))reloc_info[PTHREAD_JOIN].real_address)(th, thread_return);
  
  ifREP_NEXT();
  
  // bereken nieuwe lc
  {
    ifREC(LCt lc_other_thread = _MR_thr_LC[ TID(th) ]);
    ifREC(_MR_calc_lt(&lc_other_thread));
  }

  DBG(printf("T%d: -pthread_join ", THREADID)); DBG(PRINTLCT);

  return status;
}

/////////////////////////////////////////////////////////////////
// EXIT /////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////

void recplay_pthread_exit(void *retval){

  DBG(printf("T%d: +pthread_exit ", THREADID)); DBG(PRINTLCT);

  // beeindig draad
  (*(void(*)(void*))reloc_info[PTHREAD_EXIT].real_address)(retval);

  // not reached
}


//////////////////////////////////////////////////////////////////
// MUTEX_LOCK ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_pthread_mutex_lock(pthread_mutex_t *mutex){
  int status;
  
  DBG(printf("T%d: +pthread_mutex_lock(%x)\n", THREADID, mutex));

  ifREP_WAIT();

  // doit
  status = (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_LOCK].real_address)(mutex);

  ifREP_NEXT();

  ifREC_UPDATE_LC(mutex);

  DBG(printf("T%d: -pthread_mutex_lock(%x)\n", THREADID, mutex));

  return status;
}

//////////////////////////////////////////////////////////////////
// MUTEX_TRYLOCK ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_pthread_mutex_trylock(pthread_mutex_t *mutex){
  int status = (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_TRYLOCK].real_address)(mutex);

  DBG(printf("T%d: pthread_mutex_trylock(%x)\n", THREADID, mutex));

  // did we get the lock?
  if (status==EBUSY) return EBUSY;  // NO

  return status;

}

//////////////////////////////////////////////////////////////////
// MUTEX_UNLOCK //////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_pthread_mutex_unlock(pthread_mutex_t *mutex){
  int status;

  DBG(printf("T%d: +pthread_mutex_unlock(%x)\n", THREADID, mutex));

  ifREC_UPDATE_LC(mutex);

  ifREP_WAIT();

  status = (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_UNLOCK].real_address)(mutex);

  ifREP_NEXT();

  DBG(printf("T%d: -pthread_mutex_unlock(%x)\n", THREADID, mutex));

  return status;
}

//////////////////////////////////////////////////////////////////
// SEM_WAIT //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_sem_wait(sem_t * sem){
  int status = (*(int(*)(sem_t*))reloc_info[SEM_WAIT].real_address)(sem);

  DBG(printf("T%d: sem_wait\n", THREADID));

  return status;
}

//////////////////////////////////////////////////////////////////
// SEM_POST //////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_sem_post(sem_t * sem){

  DBG(printf("T%d: sem_post\n", THREADID));

  return (*(int(*)(sem_t*))reloc_info[SEM_POST].real_address)(sem);
}

//////////////////////////////////////////////////////////////////
// SEM_TRYWAIT ///////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_sem_trywait(sem_t * sem){
  int status = (*(int(*)(sem_t*))reloc_info[SEM_TRYWAIT].real_address)(sem);

  DBG(printf("T%d: sem_trywait\n", THREADID));

  // Hebben we de semafoor?
  if (status==EAGAIN) return EAGAIN;   // NEE

  return status;
}

//////////////////////////////////////////////////////////////////
// COND_WAIT /////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex){
  int status;

  DBG(printf("T%d: +pthread_cond_wait(%x, %x)\n", THREADID, cond, mutex));

  // switch voor interne mutex_unlock. Merk op dat we de mutex nog hebben.

  status = (*(int(*)(pthread_cond_t*, pthread_mutex_t*))reloc_info[COND_WAIT].real_address)(cond, mutex);

  // switch voor interne mutex_lock. Merk op dat we de mutex reeds terug hebben.

  DBG(printf("T%d: -pthread_cond_wait(%x, %x)\n", THREADID, cond, mutex));

  return status;
}

//////////////////////////////////////////////////////////////////
// COND_TIMEDWAIT ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

int recplay_pthread_cond_timedwait(pthread_cond_t *cond, pthread_mutex_t *mutex, 
				   const struct timespec * abstime){
  int status;

  DBG(printf("T%d: pthread_cond_timedwait\n", THREADID));

  // switch voor interne mutex_unlock. Merk op dat we de mutex nog hebben.


  status = (*(int(*)(pthread_cond_t*, pthread_mutex_t*, const struct timespec*))
	    reloc_info[COND_TIMEDWAIT].real_address)(cond, mutex, abstime);

  // switch voor interne mutex_lock. Merk op dat we de mutex reeds terug hebben, ook al was er een timeout!

  return status;
}

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

recplay_backend_init(){
  int i=0;

  fprintf(jiti_tty, ifREC("Record")ifREP("Replay")" phase\n" );
  
  if (recplay_real_pthread_create = jiti_callback("pthread_create", (unsigned)recplay_pthread_create)) {
    
    // PARALLEL PROGRAM

    // onderschep alle synchronisatiefuncties
    while (i<AANTAL_INSTR_FUNC){
      reloc_info[i].real_address = jiti_callback(reloc_info[i].function, reloc_info[i].instr_address);
      i++;
    }

    // onderschep het beeindigen v/h programma
    jiti_callback("#STOP#", (unsigned)done);

    ifREC(_MR_start_record());    
    ifREP(_MR_start_replay());

  } else {
    // SEQUENTIAL PROGRAM
  }

}

//////////////////////////////////////////////////////////////////
//INTERNE functies ///

int _MR_INTERN_mutex_lock(pthread_mutex_t * mutex){
  return reloc_info[PTHREAD_MUTEX_LOCK].real_address ?
    (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_LOCK].real_address)(mutex) :
    pthread_mutex_lock(mutex) ;
}

int _MR_INTERN_mutex_unlock(pthread_mutex_t * mutex){
  return reloc_info[PTHREAD_MUTEX_UNLOCK].real_address ? 
    (*(int(*)(pthread_mutex_t*))reloc_info[PTHREAD_MUTEX_UNLOCK].real_address)(mutex) :
    pthread_mutex_unlock(mutex) ;
}

int _MR_INTERN_sem_wait(sem_t * sem){
  return reloc_info[SEM_WAIT].real_address ?
    (*(int(*)(sem_t*))reloc_info[SEM_WAIT].real_address)(sem) :
    sem_wait(sem) ;
}

int _MR_INTERN_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex){
  return reloc_info[COND_WAIT].real_address ?
    (*(int(*)(pthread_cond_t *, pthread_mutex_t *))reloc_info[COND_WAIT].real_address)(cond, mutex) :
    pthread_cond_wait(cond, mutex) ;
}

int _MR_INTERN_cond_signal(pthread_cond_t *cond){
  return pthread_cond_signal(cond) ;        // wordt niet onderschept !!! 
}



//////////////////////////////////////////////////////////////////

INIT(recplay_backend_init);

///////////////////////

#define T_SIZE 100

struct _O_2_LCP_ {
  unsigned address;
  LCt * LCp;
} O_2_LCP[T_SIZE];

LCt * recplay_get_Object_LCP(void * address){
  int i=0, empty_spot=-1;

  while (i<T_SIZE){

    // found address
    if (O_2_LCP[i].address==(unsigned)address)  return O_2_LCP[i].LCp;

    // found empty spot
    if ((O_2_LCP[i].address==0) && (empty_spot==-1)) empty_spot=i;
    
    i++;
  }

  // not found
  //  case 1: table full
  if (empty_spot==T_SIZE) panic("O_2_LCP is full!\n");
  
  //  case 2: empty spot at i
  O_2_LCP[empty_spot].address = (unsigned)address;
  O_2_LCP[empty_spot].LCp     = calloc(1, sizeof(LCt));

  return O_2_LCP[empty_spot].LCp;
}

////

cleanup_file_name(char * naam){
  char * loper=naam;

  while (*loper++) if (*loper=='/') *loper='_';
}
