/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti-ca.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2001/10/25 13:46:01 $
 * $Log: jiti-ca.c,v $
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend analyses the code coverage. After each run of 
 * a particular program, the file <progname>.ca.log is updated to reflect
 * the code executed during this and the previous runs. 
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"
#include <stdio.h>

extern FILE * jiti_tty;

extern disassemble_info info;

static struct _blok_ {
  unsigned from;
  unsigned to;
  struct _blok_ * prev;
  struct _blok_ * next;
} * exec;


#define FIRST 0x00000000
#define LAST  0xFFFFFFFF

//extern unsigned _fini;
//extern unsigned _init;

/////////////////////////////////////

static _ca_print_empty_ranges();

static _ca_done(){
  fprintf(jiti_tty, "===== JiTI-CA: done\n");
  _ca_print_ranges();
}

/////////////////////////////////////

static _ca_doit(unsigned from, unsigned to);

static void _ca_trace_code(unsigned from, unsigned to){
  unsigned i=from;

#if 0
  fprintf(jiti_tty, "> 0x%.8x - 0x%.8x C\n", from, to);
  
  while(i<=to) {
    fprintf(jiti_tty, ">\t");
    i+=jiti_fprint_ins1(jiti_tty, 1, i);
    fprintf(jiti_tty, "\n");
  }
#endif  

  _ca_doit(from, to);
}
  
////////////////////////////////////

static _ca_doit(unsigned from, unsigned to){
  struct _blok_ * runner1, * runner2, * runner3, * new_blok;


  runner1 = exec;

  while (1) {

    runner2 = runner1->next;

    // mogelijkheid 1:    runner1.from <= from < to <= runner1.to
    if ((from>=runner1->from) && (to<=runner1->to))
      return;
    
    // mogelijkheid 2:    runner1.to+1 = from
    if (runner1->to+1==from){
      // mogelijkheid 2a: to+1<runner2.from
      if (to+1<runner2->from) {
	runner1->to = to;
	return;
      }

      // mogelijkheid 2b: to+1==runner2.from
      if (to+1==runner2->from) {
	runner1->to = runner2->to;
	runner1->next = runner2->next;
	runner2->next->prev = runner1;
	free(runner2);
	return;
      }
      
      // mogelijkheid 2c: to+1>runner2.from
      //  in dit geval moet to+1<runner2.to
      
      runner1->to = runner2->to;
      runner1->next = runner2->next;
      runner2->next->prev = runner1;
      free(runner2);
      return;

    }


    // mogelijkheid 3:    runner1.to+1 = from < to <= runner2.from
    if (runner1->to+1==from) {
      runner1->to = to;
      return;
    }

    // mogelijkheid 4:    runner1.to <= from < to <= runner2.from
    if (to<runner2->from) {
      new_blok=(struct _blok_ *)malloc(sizeof(struct _blok_));
      new_blok->from = from;
      new_blok->to   = to;
      new_blok->prev = runner1;
      new_blok->next = runner2;
      runner1->next = new_blok;
      runner2->prev = new_blok;
      return;
    }

    // mogelijkheid 5: to >= runner2->from
    runner1=runner2;
  }
}

////////////////////////////////////

int _ca_print_ranges(){
  FILE * file;
  char naam[100];
  unsigned executed=0;
  struct _blok_ * runner = exec->next;   /* [call main, halt] overslaan */

  sprintf(naam, "%s.ca.log", _jiti_naam);
  file=fopen(naam, "w");

  fprintf(jiti_tty, "Executed:\n");

  do {
    fprintf(jiti_tty, "   [%x-%x]\n", runner->from, runner->to);
    fprintf(file, "%x-%x\n", runner->from, runner->to);
    executed += runner->to-runner->from+1;
    runner=runner->next;
  } while (runner->next);
  
  
  fprintf(jiti_tty, "Total executed bytes: %d\n\n", executed);
  fprintf(stderr, "\n>>>  Total executed bytes: %d<<<\n\n", executed);
  
  fclose(file);
}

////////////////////////////////////

static _ca_print_empty_ranges(){
  unsigned i;
  struct _blok_ * runner = exec->next->next;   /* [call main, halt] overslaan */
  
  fprintf(jiti_tty, "Not executed:\n");

  do {
    fprintf(jiti_tty, "   [%x-%x]\n", runner->to+1, runner->next->from-1);

    // dissassemble
    i=runner->to+1;
    while(i<runner->next->from) {      
      fprintf(jiti_tty, ">\t%.8x: ", i);

      jiti_addr2line(i);
      
      i+=jiti_fprint_ins1(jiti_tty, 0, i);
      fprintf(jiti_tty, "\n");
    }
        
    runner=runner->next;
  } while (runner->next->from!=LAST);
}

//////////////////////////////////////

static _ca_init(){
  struct _blok_ * exec2;
  FILE * file;
  char naam[100];

  fprintf(jiti_tty, "===== JiTI-CA: Collect memory ranges accesses as code =====\n");

  jiti_callback("#STOP#", _ca_done);
  jiti_callback("#CODE#", _ca_trace_code);
  
  exec  = (struct _blok_ *)malloc(sizeof(struct _blok_));
  exec2 = (struct _blok_ *)malloc(sizeof(struct _blok_));

  exec->from = FIRST;
  exec->to   = FIRST;
  exec->prev = NULL;
  exec->next = exec2;
 
  exec2->from = LAST;
  exec2->to   = LAST;
  exec2->prev = exec;
  exec2->next = NULL;
  
  // bevolk de lijst
  sprintf(naam, "%s.ca.log", _jiti_naam);
  file=fopen(naam, "r");
  if (file) {
    unsigned from, to;
    while (!feof(file)){
      fscanf(file, "%x-%x\n", &from, &to);
      _ca_doit(from, to);
    }	
    fclose(file);
  }

}

INIT(_ca_init);
