/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti-malloc.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: ronsse $
 * $Date: 2002/02/22 09:42:53 $
 * $Log: jiti-malloc.c,v $
 * Revision 1.4.2.5  2002/02/22 09:42:53  ronsse
 * jiti-malloc.c
 *
 * Revision 1.4.2.4  2002/02/21 13:33:37  jmaebe
 * * forgot to resolve a conflict in previous merge
 *
 * Revision 1.4.2.3  2002/02/21 13:13:06  jmaebe
 * * The stack is no longer part of the malloced-regions. This was a bug that showed
 *   up in mem5 and mem6 (= merged 1.6 - 1.7)
 *
 * Revision 1.4.2.2  2002/02/20 08:17:04  jmaebe
 * + support for dynlibs (untested)
 * * The top of the stack is calculated instead of guessed to be 0xC0000000
 *   (merged diffs between rev 1.5 and 1.6 from main branch)
 *
 * Revision 1.4.2.1  2002/02/15 11:25:33  jmaebe
 * + Changed scanf pattern to enable the scanning of libstdc++ (merged)
 *
 * Revision 1.4  2001/11/12 10:13:54  ronsse
 * Added 1 to size calculation
 *
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend checks an execution for all kinds of memory faults,
 * such as:
 * 	-reading/writing outside an array
 * 	-freeing memory that is aleady freed.
 * 	-..
 *
 * It checks all memory accesses against
 * 	-the memory that is allocated statically
 * 	-the memory that is allocated dynamically, e.g. usign malloc(), mmap(), ...
 * 	-the runtime stack.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



//#define DEBUG

/*

 Implementeert:

   malloc/calloc/realloc/alloca/free
   mmap/munmap/mremap

 TODO: brk/sbrk ??

 MERK OP: void * alloca(size) is geen functie! GCC genereert het volgende:

     addl   -size,%esp
     movl   %esp,%eax

*/

#include <unistd.h>
#include <sys/mman.h>  

#include "jiti.h"


unsigned real_malloc;
#define MALLOC(SIZE) (real_malloc ? (*(void*(*)(size_t))real_malloc)(SIZE) : malloc(SIZE))

unsigned real_free;
#define FREE(PTR) (real_free ? (*(void(*)(void*))real_free)(PTR) : free(PTR))

unsigned real_calloc;
#define CALLOC(NMEMB, SIZE)  (*(void*(*)(size_t,size_t))real_calloc)(NMEMB, SIZE)

unsigned real_realloc;
#define REALLOC(PTR, SIZE)   (*(void*(*)(void*, size_t))real_realloc)(PTR, SIZE)

unsigned real_mmap;
#define MMAP(START, LENGTH, PROT, FLAGS, FD, OFFSET) \
 ((void*(*)(void*, size_t, int, int, int, off_t))real_mmap)(START, LENGTH, PROT, FLAGS, FD, OFFSET)

unsigned real_munmap;
#define MUNMAP(START, LENGTH)  ((int(*)(void*, size_t))real_munmap)(START, LENGTH)

unsigned real_mremap;
#define MREMAP(OLD_ADDRESS, OLD_SIZE, NEW_SIZE, FLAGS) \
  ((void*(*)(void *, size_t, size_t, unsigned long))real_mremap)\
  (OLD_ADDRESS, OLD_SIZE, NEW_SIZE, FLAGS)

unsigned data_code_begin, data_code_einde, stapel_einde;

#define NO_FREE 1 // geheugen dat niet gefreed/geunmapt moet worden (bv. bibliotheken)

/////////////////////////////////

// bepaal het begin v/d stapel (3 stackframes terugkeren!)
static unsigned stapel_begin(){
  asm("movl (%ebp),%eax");
  asm("movl (%eax),%eax"); 
  asm("movl (%eax),%eax"); 
  asm("addl $24,%eax"); 
}

/////////////////////////////////

static unsigned ip2source(unsigned ip){
  if (ip>0x40000000) return 0; // 't is gedaan, we zijn in jiti gearriveerd
  printf("\n\t\t"); 
  jiti_addr2line(ip);
  return 1;  // op naar het volgende frame
}

/////////////////////////////////

// informatie over 1 geallokeerd gebied geheugen
typedef struct _malloc_info_ {
  unsigned start;
  unsigned stop;
  unsigned type;
  unsigned ip;
  struct _malloc_info_ * next;
} _malloc_info_;

// gelinkte lijst met info over alle geallokeerde gebieden geheugen (momentopname)
_malloc_info_ * __malloc_info=NULL;

///////////////////////////////////

static unsigned count_malloced_regions(){
  unsigned count=0;
 _malloc_info_ * runner = __malloc_info;
  while (runner){
    if (runner->type!=NO_FREE) count++;
    runner = runner->next;
  }
  return count;
}

///////////////////////////////////

static size_t print_malloced_regions(){
  size_t size=0;
  _malloc_info_ * runner = __malloc_info;
  while (runner){
    if (runner->type!=NO_FREE){
      printf("\n[0x%.8x -> 0x%.8x : %d bytes]\n", runner->start, runner->stop, 
	     runner->stop-runner->start+1);
      ip2source(runner->ip);
      size+=runner->stop-runner->start+1;
    }
    runner = runner->next;
  }
  return size;
}

//////////////////////////////////

static unsigned access_to_malloced_region(unsigned start, unsigned stop){
  _malloc_info_ * runner = __malloc_info;
  stop--;
  while (runner){
    if ((start>=runner->start) && (stop<=runner->stop))  return 1;
    runner = runner->next;
  }
  return 0;
}

///////////////////////////////////

static void add_to_malloced_regions(unsigned begin, unsigned end, unsigned type, unsigned ip){
  _malloc_info_ * new_malloc_info=MALLOC(sizeof(_malloc_info_));
  end--;
  new_malloc_info->start = begin;
  new_malloc_info->stop  = end;
  new_malloc_info->type  = type;
  if (ip-delta <= end)
    new_malloc_info->ip  = ip-delta-5;
  else
    new_malloc_info->ip  = ip-ddelta-5;
  new_malloc_info->next  = __malloc_info;
  __malloc_info = new_malloc_info;  
}

//////////////////////////////////

static size_t remove_from_malloced_regions(unsigned begin){
  size_t size=0;
  void * temp;
 _malloc_info_ ** runner = &__malloc_info;
 
 while (*runner){
   if ((*runner)->start==begin) { // OK, gevonden
     size=(*runner)->stop-(*runner)->start+1;
     temp = *runner;
     (*runner)=(*runner)->next;
     FREE(temp);
     return size;
   }
   runner=&((*runner)->next);
 }
 
 // niet gevonden
 return 0;
}

///////////////////////////////////

static int remove_from_malloced_regionsPART(unsigned begin, unsigned end){
  size_t size=0;
  void * temp;
  int OK=0;
  _malloc_info_ ** runner = &__malloc_info;
  
  end--;

  while (*runner){

    //printf("begin=%x end=%x start=%x stop=%x\n", begin, end, (*runner)->start, (*runner)->stop);
    
    // geen overlapping
    if (begin >= (*runner)->stop)  goto NEXT;
    if (end   <= (*runner)->start) goto NEXT;
    
    OK|=1;

    // totale overlapping => gooi blok weg
    if ((begin <= (*runner)->start) && (end >= (*runner)->stop)) {
      temp = *runner;
      (*runner)=(*runner)->next;
      FREE(temp);
      goto NEXT;
    }

    // Nu knippen we steeds een blok [begin, end] uit [start, stop] => mogelijk
    // krijgen we twee blokken: [start, begin] en [end, stop].

    // mogelijkheid 1: start < begin < stop <= end  
    if (((*runner)->start<begin) && ((*runner)->stop<=end)) {
      (*runner)->stop=begin;  // we houden [start, begin] over
      goto NEXT;
    }

    // mogelijkheid 2: begin <= start < end < stop
    if ((*runner)->start>=begin){
      (*runner)->start=end;  // we houden [end, stop] over
      goto NEXT;
    }
  
    // mogelijkheid 3: start < begin < end < stop
    // we houden [start, begin] en [end, stop] over
    {
      _malloc_info_ * new_malloc_info=MALLOC(sizeof(_malloc_info_));
      new_malloc_info->start = end;
      new_malloc_info->stop  = (*runner)->stop;
      new_malloc_info->next  = (*runner)->next;
      (*runner)->stop        = begin;
      (*runner)->next        = new_malloc_info;
    }
    goto NEXT;

  NEXT: 
    if (!(*runner)) return OK;

    runner=&((*runner)->next);   
  }
  
  return OK;
}

///////////////////////////////////

static void * _malloc_our_malloc(size_t size){
  void * result;  
  
  result = MALLOC(size);
  DBG(printf("malloc(%d) @ 0x%x\n", size, result));

  if (result) add_to_malloced_regions((unsigned)result, (unsigned)(result+size), 0, get_ip());

  return result;
}

//////////////////////////////////

static void * _malloc_our_calloc(size_t nmemb, size_t size){
  void * result;

  result=CALLOC(nmemb, size);
  DBG(printf("calloc(%d, %d) @ 0x%x\n", nmemb, size, result));

  if (result) add_to_malloced_regions((unsigned)result, (unsigned)(result+nmemb*size), 0, get_ip());

  return result;
}

///////////////////////////////////

static void _malloc_our_free(void *ptr){
  size_t size;
  unsigned ip;

  if (!ptr) return;   // free(NULL) is steeds toegestaan

  if (size=remove_from_malloced_regions((unsigned)ptr)){
    DBG(printf("free(0x%x) released %d bytes\n", ptr, size));
    FREE(ptr);
    return;
  } else {
    printf("\tFreeing an unallocated block of memory @ 0x%x!!\n\tSource:\n\t", ptr);
    ip2source(((ip = get_ip())<=end?ip-delta:ip-ddelta)-5);
    _exit(-1);
  }

}

//////////////////////////////////

static void * _malloc_our_realloc(void * ptr, size_t size){
  void * result;
  size_t original_size;
  unsigned ip;

  if (!ptr) {                // -> malloc()
    result=REALLOC(ptr, size);
    DBG(printf("realloc(0x%x, %d) @ 0x%x\n", ptr, size, result));
    if (result) add_to_malloced_regions((unsigned)result, (unsigned)(result+size), 0, get_ip());
    return result;
  } else if (!size) {        // -> free()
    if (remove_from_malloced_regions((unsigned)ptr)) {
      result=REALLOC(ptr, size);
      DBG(printf("realloc(0x%x, %d) @ 0x%x\n", ptr, size, result));
      return;
    } else {
      DBG(printf("realloc() terwijl %x niet geallokeerd is!!\n", ptr));
      ip2source(((ip = get_ip())<=end?ip-delta:ip-ddelta)-5);
      _exit(-1);
    }      
  } else {                   // -> normaal
    if (original_size=remove_from_malloced_regions((unsigned)ptr)) {
      result=REALLOC(ptr, size);
      DBG(printf("realloc(0x%x, %d) @ 0x%x\n", ptr, size, result));
    } else {
      DBG(printf("realloc() terwijl %x niet geallokeerd is!!\n", ptr));
      ip2source(((ip = get_ip())<=end?ip-delta:ip-ddelta)-5);
      _exit(-1);
    }      
    if (result)
      add_to_malloced_regions((unsigned)result, (unsigned)(result+size), 0, get_ip());
    else // niet geslaagd => oude situatie herstellen
      add_to_malloced_regions((unsigned)result, (unsigned)(result+original_size), 0, get_ip());
  }  
  return result;
}

//////////////////////////////////

static void * _malloc_our_mmap(void *start, size_t length, int prot , int flags, int fd, off_t offset){
  void * result;
  result=MMAP(start, length, prot, flags, fd, offset);

  DBG(printf("mmap(0x%x, %d, %d, %d, %d, %d) @ 0x%x\n", 
	     start, length, prot, flags, fd, offset, result));
  
  if (result!=MAP_FAILED) add_to_malloced_regions((unsigned)result, (unsigned)(result+length), 
						  0, get_ip());
  
  return result;
}

//////////////////////////////////

static int _malloc_our_munmap(void *start, size_t length){
  int result;
  unsigned size,ip;

  if (remove_from_malloced_regionsPART((unsigned)start, (unsigned)(start+length))){
    DBG(printf("munmap(0x%x) released %d bytes (?)\n", start, length));
    result=MUNMAP(start, length);
    if (result) printf("munmap(%x, %d) failed!\n", start, length);
    return result;
  } else {
    printf("Unmapping a block (@0x%x) that was never mapped!!\n", start);
    ip2source(((ip = get_ip())<=end?ip-delta:ip-ddelta)-5);
    _exit(-1);
  }
}

//////////////////////////////////

static void * _malloc_our_mremap(void * old_address, size_t old_size , size_t new_size, unsigned long flags){
  void * result;

  remove_from_malloced_regionsPART((unsigned)old_address,(unsigned) old_size);
  result=MREMAP(old_address, old_size, new_size, flags);
  if (result!=(void*)-1) add_to_malloced_regions((unsigned)result,(unsigned) new_size, 0, get_ip());
  return result;
}

//////////////////////////////////


static void _malloc_memory_access(unsigned ip, unsigned type, unsigned len, unsigned address){

  DBG(printf("MEM: %.8x/%d/%d/%.8x [%x]\n", ip, type, len, address, stapel_begin()));

  //    Mogelijkheid 1:  stapel
  if ((address>=stapel_begin()) && (address+len<=stapel_einde)) return;

  //    Mogelijkheid 2:  DS of CS
  if ((address>=data_code_begin) && (address+len<=data_code_einde)) return;;

  //    Mogelijkheid 3:  malloc/mmap-ed
  if (access_to_malloced_region(address, address+len)) return;

  // GEEN TOEGANG
  printf("\n\n\tInvalid memory access @ 0x%x using instruction @ 0x%x!\n\n"
	 "\tThe instruction is:\n\t\t", 
	 address, ip);
  jiti_fprint_ins1(stdout, 0, ip);
  printf("\n\n\tThe source code is:\n\t\t");
  jiti_addr2line(ip);
  printf("\n\tThe call chain is:");
  print_my_stack(ip2source, 4);
  _exit(-1);


}

//////////////////////////////////


static _malloc_program_done(){
  fprintf(jiti_tty, "===== JITI-MALLOC: done ======\n");

  // zijn er malloc()s die niet gefree()d zijn?
  if (count_malloced_regions()){
    printf("\n\nThe following memory regions were dynamically allocated, but they were never released:");
    printf("\nTotal size: %d bytes\n\n", print_malloced_regions());
  }

}

//////////////////////////////////

static _malloc_backend_init2(){

  // alle tot nu toe verzamelde malloc-info weggooien
  __malloc_info = NULL;  // merk op: geheugenlek !!!
  
  // bepaal wat nu reeds gemapt is -> voeg toe d.m.v. add_to_malloced_regions
  {
    char maps[256];
    FILE * maps_file;
    unsigned from, to;

    maps_file=fopen("/proc/self/maps", "r");
    while (EOF!=fscanf(maps_file, "%x-%x %[a-zA-Z0-9 :/._-+]\n", &from, &to, &maps)){
      fprintf(jiti_tty, "[%x-%x] %s\n", from, to, maps);
      if (to==stapel_einde) break;

      if ((from>=0x40000000) && (from<=0xbfffb000)) // enkel libs, niet de stack of .text
	add_to_malloced_regions(from, to, NO_FREE, 0);
    }
    fclose(maps_file);

  }
  
  DBG(printf("MALLOC: real start\n"));


}

////////////////////////////////////

static _malloc_backend_init(){
  int dummy;

  fprintf(jiti_tty, "===== JiTI-MALLOC: check memory accesses + malloc(),...  ======\n");

  // bereken begin en einde van data en code segment
  data_code_begin = 0x8048000;
  FREE((void*)(data_code_einde=(unsigned)MALLOC(1)));
  
  // Finding the top of the stack is based on the fact that this backend is
  // using the top of the stack as it is almost directly called by init().
  stapel_einde = ((unsigned)(&dummy) & 0xFFFF0000) + 0x00010000;

  // zorg dat we alle `interessante' gebeurtenissen onderscheppen
  real_malloc      = jiti_callback("malloc",  (unsigned)_malloc_our_malloc);  
  real_free        = jiti_callback("free",    (unsigned)_malloc_our_free);
  real_calloc      = jiti_callback("calloc",  (unsigned)_malloc_our_calloc);  
  real_realloc     = jiti_callback("realloc", (unsigned)_malloc_our_realloc);
  real_mmap        = jiti_callback("mmap",    (unsigned)_malloc_our_mmap);
  real_munmap      = jiti_callback("munmap",  (unsigned)_malloc_our_munmap);

  jiti_callback("#STOP#",  (unsigned)_malloc_program_done);
  jiti_callback("#START#", (unsigned)_malloc_backend_init2);  
  jiti_callback("#DATA#",  (unsigned)_malloc_memory_access);

}

INIT(_malloc_backend_init);
