/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti-mem.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/02/26 13:09:15 $
 * $Log: jiti-mem.c,v $
 * Revision 1.2.2.1  2002/02/26 13:09:15  jmaebe
 * * _mem_doit() now concatenates adjoining ranges in all possible cases
 * - removed unused variable
 *
 * Revision 1.2  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend calculates ranges of memory read/written during
 * an execution.
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "jiti.h"
#include <stdio.h>

extern FILE * jiti_tty;

#include "dis-asm.h"
extern disassemble_info info;

static struct _blok_ {
  unsigned from;
  unsigned to;
  struct _blok_ * prev;
  struct _blok_ * next;
} * _mem_exec;

static unsigned _mem_aantal_Sb=0, _mem_aantal_Lb=0;
static unsigned _mem_aantal_S=0, _mem_aantal_L=0;


static unsigned _mem_aantal; 

#define FIRST 0
#define LAST 0xFFFFFFFF

////////////////////////////////////

static _mem_doit(unsigned from, unsigned to){
  struct _blok_ * runner1, * runner2, * runner3, * new_blok;

  runner1 = _mem_exec;

  while (1) {

    runner2 = runner1->next;

    // case 1:    runner1.from <= from < to <= runner1.to
    if ((from>=runner1->from) && (to<=runner1->to))
      return;
    
    // case 2:    from <= runner1.to+1
    if (from <= runner1->to+1) {

      // from < runner1->from => adjust runner1->from
      if (from < runner1->from)
        runner1->from = from;

      // case 2a: to+1<runner2.from
      if (to+1<runner2->from) {
        runner1->to = to;
        return;
      }

      // case 2b: to+1<=runner2.to
      if (to+1<=runner2->to) {
        runner1->to = runner2->to;
        runner1->next = runner2->next;
        runner2->next->prev = runner1;
        free(runner2);
        return;
      }
      
      // case 2c: to+1>runner2.to
      runner1->to = to;
      runner1->next = runner2->next;
      runner2->next->prev = runner1;
      free(runner2);
      return;
    }

    // case 3:    runner1.to+1 < from < to < to+1 < runner2.from
    if (to+1<runner2->from) {
      new_blok=(struct _blok_ *)malloc(sizeof(struct _blok_));
      new_blok->from = from;
      new_blok->to   = to;
      new_blok->prev = runner1;
      new_blok->next = runner2;
      runner1->next = new_blok;
      runner2->prev = new_blok;
      return;
    }

    // case 5: (from > runner1.to+1) && (to+1 >= runner2->from)
    runner1=runner2;
  }
}

////////////////////////////////////

static _mem_print_ranges(){
  unsigned accessed=0;
  struct _blok_ * runner = _mem_exec->next;
  
  fprintf(jiti_tty, "\n===== JiTI-MEM: done\nAccessed:\n");
  
  do {
    fprintf(jiti_tty, "   [0x%.8x-0x%.8x] %8d\n", 
	    runner->from, 
	    runner->to, 
	    runner->to-runner->from+1);
    accessed += runner->to-runner->from+1;
    runner=runner->next;
  } while (runner->next);
  
  fprintf(jiti_tty, "\nTotal bytes:               %8d\n", accessed);  
  fprintf(jiti_tty, "%8d read operation read        %8d bytes (%2.2f bytes/op)\n",
          _mem_aantal_L, _mem_aantal_Lb,  (float)_mem_aantal_Lb/(float)_mem_aantal_L);
  fprintf(jiti_tty, "%8d store operations stored    %8d bytes (%2.2f bytes/op)\n",
          _mem_aantal_S, _mem_aantal_Sb,  (float)_mem_aantal_Sb/(float)_mem_aantal_S);
  fprintf(jiti_tty, "%8d operations accessed        %8d bytes (%2.2f bytes/op)\n",
          _mem_aantal, _mem_aantal_Sb+_mem_aantal_Lb, 
	  (float)(_mem_aantal_Sb+_mem_aantal_Lb)/(float)_mem_aantal); 
  
}

////////////////////////////////////

static _mem_trace(unsigned ip, unsigned type, unsigned len, unsigned address){

  _mem_aantal++;

  switch (type){
  case LOAD:   _mem_aantal_Lb += len, _mem_aantal_L++;  break;
  case STORE:  _mem_aantal_Sb += len, _mem_aantal_S++;  break;
  case MODIFY: _mem_aantal_Lb += len, _mem_aantal_L++, _mem_aantal_Sb += len, _mem_aantal_S++;  break;
  }

  _mem_doit(address, address+len-1);

}  

////////////////////////////////////

static _mem_init(){
  struct _blok_ * exec2;

  fprintf(jiti_tty, "===== JiTI-MEM: Collect memory range accesses as data =====\n");

  jiti_callback("#STOP#", (unsigned)_mem_print_ranges);
  jiti_callback("#DATA#",  (unsigned)_mem_trace);
   
  _mem_exec  = (struct _blok_ *)malloc(sizeof(struct _blok_));
  exec2 = (struct _blok_ *)malloc(sizeof(struct _blok_));

  _mem_exec->from = FIRST;
  _mem_exec->to   = FIRST;
  _mem_exec->prev = NULL;
  _mem_exec->next = exec2;
 
  exec2->from = LAST;
  exec2->to   = LAST;
  exec2->prev = _mem_exec;
  exec2->next = NULL;

}

INIT(_mem_init);
