/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  jiti-mem.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/03/12 08:30:27 $
 * $Log: trace-mem.c,v $
 * Revision 1.2.2.4  2002/03/12 08:30:27  jmaebe
 * * don't trace some illegal addresses (they're not really illegal, but they
 *   use the gs selector which isn't supported by JiTI yet)
 *
 * Revision 1.2.2.3  2002/03/01 10:26:12  ronsse
 * Fancier output
 *
 * Revision 1.2.2.2  2002/02/28 15:54:04  jmaebe
 * * use spinlocks instead of pthread_mutex_lock(), because the latter is not
 *   re-entrant (so we hang if the trace_data_par() function is called while
 *   we are instrumenting pthread_mutex_lock())
 *
 * Revision 1.2.2.1  2002/02/27 10:50:48  jmaebe
 * * intialize thge mutex used when ijsntrumenting parallel programs before
 *   using it
 * + write the (previous) value at the instrumented memory addresses (current
 *   value in case of loads, previous value in case of stores/modications)
 *
 * Revision 1.2  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * This JiTI86 backend generates a list of memory accesses perfromed
 * during a program run. For each memory access, the following infromation is shown:
 * 	-the type of operation: S(tore), L(oad) or X(=load+store)
 * 	-the identifier of the thread perfroming the operation (for parallel programs)
 * 	-the address of the instruction performing the memory operation
 * 	-the memory address used
 * 	-the number of bytes read/written
 * 	-the assembly instruction
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"

char c_type[3]={'L', 'S', 'X'};

// uitvoerformaat voor het data-adres
#define OUTPUT_FORMAT "%.8x"    // hexadeximaal
#define _OUTPUT_FORMAT "%u"      // decimaal


/////////////////////////////////////

static void trace_data(unsigned tid, unsigned ip, unsigned type, unsigned len, unsigned address){

  if ((unsigned)address < (unsigned)0x8000000) {
	fprintf(jiti_tty, "Accessing illegal address 0x%x from 0x%x (thread %d), not tracing\n",address,ip,tid);
	return;
  }
  switch (len) {
    case 1: {
	      unsigned char kar=GET_CHAR(address);
	      fprintf(jiti_tty, "%c %d 0x%.8x: " OUTPUT_FORMAT " %4d 0x%.2x   ", 
		  c_type[type], tid, ip, address, len, kar); 
	      if ((kar>31) && (kar<127)) 
		fprintf(jiti_tty, "[%c] ", kar);
	      else 
		fprintf(jiti_tty, "    ", kar);
	    }
	    break;
    case 2: 
	    fprintf(jiti_tty, "%c 1 0x%.8x: " OUTPUT_FORMAT " %4d 0x%.4x     ", 
		c_type[type], ip, address, len, GET_SHORT(address)); 
	    break;
    case 4: 
	    fprintf(jiti_tty, "%c 1 0x%.8x: " OUTPUT_FORMAT " %4d 0x%.8x ", 
		c_type[type], ip, address, len, GET_LONG(address));
	    break;
    default:
	    fprintf(jiti_tty, "%c 1 0x%.8x: " OUTPUT_FORMAT " %4d            ", 
		c_type[type], ip, address, len);
	    break;
  }

  jiti_fprint_ins1(jiti_tty, 0, ip);
  fprintf(jiti_tty, "\n");

}

/////////////////////////////////////

static void trace_data_seq(unsigned ip, unsigned type, unsigned len, unsigned address){
  trace_data(1, ip, type, len, address);
}

static void trace_data_par(unsigned ip, unsigned type, unsigned len, unsigned address){
  unsigned p=THREADID;
  static unsigned trace_data_par_lock = 0;
  
  __wait_and_set((unsigned)&trace_data_par_lock);
  trace_data(p, ip, type, len, address);
  trace_data_par_lock = 0;
}

/////////////////////////////////////

static init_trace_data(){

  fprintf(jiti_tty, "===== MEM_INIT: Trace all data accesses ======\n");
  jiti_callback("#DATA#", (unsigned)(jiti_parallel_program() ? trace_data_par : trace_data_seq));

}

/////////////////////////////////////

INIT(init_trace_data);

