/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *	dynlib.c 
 *
 *	Copyright (C) 2002 by Jonas Maebe, Ghent University, Belgium
 *
 *	This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.	
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/03/12 09:19:34 $
 * $Log: dynlib.c,v $
 * Revision 1.1.2.5  2002/03/12 09:19:34  jmaebe
 * * fixed detection of libpthread. All parallel test programs now run fine
 *   when instrumented with any combination of the opt, trace-mem and
 *   trace-code backends!!! Hell seems to have finally frozen over :)
 *
 * Revision 1.1.2.4  2002/03/11 21:36:53  jmaebe
 * * forgot to commit, see revision 1.7.2.32 of instrument.c for log message
 *
 * Revision 1.1.2.3  2002/02/27 10:31:23  jmaebe
 * * dynamically detect stack region like in jiti-malloc
 *
 * Revision 1.1.2.2  2002/02/21 12:27:39  jmaebe
 * * the gathering of the memory regions occupied by dynlibs is now done from
 *   /proc/self/maps (like in jiti-malloc) -> ld.so is now also found.
 *
 * Revision 1.1.2.1	 2002/02/12 16:47:20	jmaebe
 * + initial implementation of instrumenting of dynamically loaded libraries,
 *	 but not yet working (even main(void) {} crashes :). Also, some backends
 *	 which have to be modified before they can work with this new feature
 *	 haven't been adapted yet. Compiler with -DINSTRUMENT_DYNLIBS to include
 *	 the functionality (it's even required currently, because there's a
 *	 reference to ddelta (dynlib delta) in handlers.s which can't be disabled
 *	 using a conditional. Method used: separate clone for the dynlibs (thanks
 *	 for the idea, Michiel :)
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <stdio.h>
#include "jiti.h"
#include "dynlib.h"

/********************* DYNLIB info reader	 *******************/

///////////////////////////////

// Base and size of memory region occupied by *all* dynamic libraries used by
// the program

size_t _dynlibs_base = 0;
size_t _dynlibs_size = 0;

unsigned libpthread_base = 0;
unsigned libpthread_top = 0;

void get_dynlibs_info(void) {
	FILE *maps_file;
	unsigned stack_end;
	size_t templow, temphigh;
	char permissions[5];
	char mapsline[256], libname[256], *libnameptr, *slashpos;

	// write the libs/memory lay-out to the log
	maps_file=fopen("/proc/self/maps", "r");
	while (NULL != fgets(mapsline,sizeof(mapsline),maps_file)) {
		fprintf(jiti_tty,"%s",mapsline);
	}
	fclose(maps_file);

	stack_end = ((unsigned)(&maps_file) & 0xFFFF0000) + 0x00010000;

	// locate the part of memory occupied by the dynlibs
	maps_file=fopen("/proc/self/maps", "r");
	while (EOF!=fscanf(maps_file, "%x-%x %4s", &templow, &temphigh, permissions)){
		//DBG(printf("[%x-%x, %s:]\n", templow, temphigh, permissions));
		// only get the parts of memory that are executable
		if ((permissions[2] == 'x') &&
		    (templow>=0x40000000) && (temphigh<stack_end)) { // just the libs, no stack or .text
			if ((_dynlibs_size == 0) ||
			    (templow < _dynlibs_base))
				_dynlibs_base = templow;
			if (temphigh - _dynlibs_base > _dynlibs_size)
				_dynlibs_size = temphigh - _dynlibs_base;

			// check if we find libpthread
			fgets(libname,sizeof(libname),maps_file);
			libnameptr = &libname[0];
			while ((slashpos = strchr(libnameptr,'/')) != NULL) {
				libnameptr = slashpos+1;
			}
			if (strncmp(libnameptr,"libpthread",strlen("libpthread")) == 0) {
				libpthread_base = templow;
				libpthread_top = temphigh;
				DBG(fprintf(jiti_tty,"libpthread found at 0x%x with size 0x%x\n",libpthread_base,libpthread_top-libpthread_base));
			}
		}
		else
			// read the rest of the line so we can continue with the
			// next entry
			if (EOF == fscanf(maps_file,"%*[a-zA-Z0-9 :/._-+]\n"))
		  		break;
	}
	fclose(maps_file);

	DBG(fprintf(jiti_tty,"\nTotal: base = 0x%x, size = 0x%x (%u)\n",
	            _dynlibs_base,_dynlibs_size,_dynlibs_size);)
}

//////////////////////////////
