/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *  par.c 
 *
 *  Copyright (C) 2001 by Michiel Ronsse, Ghent University, Belgium
 *
 *  This file is part of JiTI86 (http://www.elis.rug.ac.be/~ronsse/jiti/)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 * $Author: jmaebe $
 * $Date: 2002/03/11 21:21:59 $
 * $Log: par.c,v $
 * Revision 1.4.2.3  2002/03/11 21:21:59  jmaebe
 * * MAKE_LINK is now no longer used while generating code, since it also
 *   caused problems when using MT code (thread ! is instrumenting, generats a
 *   link to its instrumented code which isn't finished yet, thread 2 returns
 *   from a function or jumps/calls to the linked address and starts executing
 *   the unfinished instrumented code) -> changed to SCHEDULE_MAKE_LINK() and
 *   committed after the current block is finished
 * * LOOKUP_CLONE2INSTR() once again uses the jumps in the clone to determine
 *   targets
 * * libpthread is no longer instrumented because it disables all signals
 *   (using syscall 0xb3) in some routines, causing JiTI to hang (since its
 *   breakpoints aren't processed anymore, it can't continue instrumenting and
 *   the program can't continue executing)
 * * made jcc_taken_handler() thread-safe: the MAKE_LINK() there was not safe
 *   since it consists of 2 PUT_****() operations, so another thread could be
 *   executing the original call while it was being replaced with a JMP. At
 *   the same time, PATCH_INFO was made obsolete by doing the patching inside
 *   jcc_taken_handler (and that should be more or less thread safe, since it
 *   replaces exactly one int) -> GEBRUIK_PATCH_INFO is always set to 0 in
 *   jiti.c, can possibly be removed.
 * + infrastructure for intercepting syscalls and modifying the values passed
 *   to the kernel, not used anymore (I used it to debug the syscall 0xb3
 *   problem)
 * * separate assembler handlers for adjusting the target of a jmp or call
 *   mem/reg when instrumenting parallel programs (the call handler checks
 *   for targets in libpthread, the jump handler doesn't). The common handler
 *   is still used if we aren't instrumenting a parallel program (and that
 *   one obviously doesn't include the pthread checks)
 * * the memory layout is now always written to the log
 * - commented out some unused global variables in instrument.c
 * * some cosmetic changes and translations
 * * with the above changes, ploop now always works (on a dual proc machine)
 *   with the opt and trace-mem backends). ptest now always crashed for some
 *   reason however. If you use trace-code, you also almost always get crashes
 *   with MT programs.
 * And that's all for today, folks!
 *
 * Revision 1.4.2.2  2002/02/22 14:20:03  jmaebe
 * - removed lock in jiti_pthread_create again since it causes a deadlock now
 *   that dynlibs are instrumented as well
 *
 * Revision 1.4.2.1  2002/02/21 15:16:36  jmaebe
 * - removed adding of delta to start routine (since pthread_create is
 *   instrumented as well, this will be done automatically)
 *
 * Revision 1.4  2001/10/29 13:53:56  ronsse
 * JiTI is locked using jiti_lock() during the creation of a new thread. This seems
 * to remove the SEGV faults.
 *
 * Revision 1.3  2001/10/25 13:46:01  ronsse
 * Initial commit
 *
 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "jiti.h"
#include <semaphore.h>


extern int _jiti_par_exec, _jiti_par_program;

int (*real_pthread_create)(pthread_t *, const pthread_attr_t *, void * (*)(void *), void *);

///////////////////////

int jiti_pthread_create(pthread_t * thread, const pthread_attr_t * attr, 
		void * (*start_routine)(void *), void * arg){
  int status;	

//  jiti_lock();
  _jiti_par_exec=1; // de parallelle uitvoering is gestart!
  
  // Merk op: dit werkt met en zonder cloon!   
  start_routine += (start_routine > end) ? ddelta : delta;
  status = (*real_pthread_create)(thread, attr, start_routine, arg);
//  jiti_unlock();

  return status;
}

///////////////////////

void par_init(){
  if ((real_pthread_create = (int(*)(pthread_t *, const pthread_attr_t *, void * (*)(void *), void *))
       jiti_callback("pthread_create", (unsigned)jiti_pthread_create))) {
    _jiti_par_program=1;
    DBG(fprintf(jiti_tty, "PAR\n"));
  } else {
    DBG(fprintf(jiti_tty, "SEQ\n"));
    _jiti_par_program=0;    
  }
}

//////////////////////
